from datetime import datetime
from typing import Optional

from flask import flash, redirect, render_template, Response, Blueprint
from flask.typing import ResponseReturnValue
from flask_wtf import FlaskForm
from sqlalchemy import exc
from wtforms import StringField, SelectField, SubmitField
from flask_wtf.file import FileField
from wtforms.validators import DataRequired

from app.extensions import db
from app.models.base import Group
from app.models.onions import Onion
from app.portal.util import response_404, view_lifecycle

bp = Blueprint("onion", __name__)


class EditOnionForm(FlaskForm):  # type: ignore
    domain_name = StringField('Domain Name', validators=[DataRequired()])
    description = StringField('Description', validators=[DataRequired()])
    tls_private_key = FileField('TLS Private Key (PEM format)',
                                description="If no file is submitted, the TLS key will remain unchanged.")
    tls_public_key = FileField('TLS Certificate (PEM format)',
                               description="If no file is submitted, the TLS certificate will remain unchanged.")
    group = SelectField('Group', validators=[DataRequired()])
    submit = SubmitField('Save Changes')


@bp.route("/new", methods=['GET', 'POST'])
@bp.route("/new/<group_id>", methods=['GET', 'POST'])
def onion_new(group_id: Optional[int] = None) -> ResponseReturnValue:
    return redirect("/ui/web/onions/new")


@bp.route('/edit/<onion_id>', methods=['GET', 'POST'])
def onion_edit(onion_id: int) -> ResponseReturnValue:
    onion: Optional[Onion] = Onion.query.filter(Onion.id == onion_id).first()
    if onion is None:
        return Response(render_template("error.html.j2",
                                        section="onion",
                                        header="404 Onion Not Found",
                                        message="The requested onion service could not be found."),
                        status=404)
    form = EditOnionForm(group=onion.group_id,
                         domain_name=onion.domain_name,
                         description=onion.description)
    form.group.choices = [(x.id, x.group_name) for x in Group.query.all()]
    if form.validate_on_submit():
        onion.group_id = form.group.data
        onion.description = form.description.data
        onion.domain_name = form.domain_name.data
        for at in [
            "tls_private_key",
            "tls_public_key"
        ]:
            if getattr(form, at).data is not None:
                # Don't clear the key if no key is uploaded
                setattr(onion, at, getattr(form, at).data.read())
        onion.updated = datetime.utcnow()
        try:
            db.session.commit()
            flash("Saved changes to group.", "success")
        except exc.SQLAlchemyError:
            flash("An error occurred saving the changes to the group.", "danger")
    return render_template("onion.html.j2",
                           section="onion",
                           onion=onion, form=form)


@bp.route("/list")
def onion_list() -> ResponseReturnValue:
    return redirect("/ui/web/onions")


@bp.route("/destroy/<onion_id>", methods=['GET', 'POST'])
def onion_destroy(onion_id: str) -> ResponseReturnValue:
    onion = Onion.query.filter(Onion.id == int(onion_id), Onion.destroyed.is_(None)).first()
    if onion is None:
        return response_404("The requested onion service could not be found.")
    return view_lifecycle(
        header=f"Destroy onion service {onion.onion_name}",
        message=onion.description,
        success_message="Successfully removed onion service.",
        success_view="portal.onion.onion_list",
        section="onion",
        resource=onion,
        action="destroy"
    )
